use super::{
    VerbContext, UserVerb, UserVerbRef, UResult,
    CommandHandlingError::UserError
};
use async_trait::async_trait;
use ansi::ansi;
use phf::phf_map;

static ALWAYS_HELP_PAGES: phf::Map<&'static str, &'static str> = phf_map! {
    "<topicname>" =>
        ansi!("You are supposed to replace <lt>topicname> with the topic you want \
         to learn about. Example:\n\
         \t<bold>help register<reset> will tell you about the register command.")
};

static UNREGISTERED_HELP_PAGES: phf::Map<&'static str, &'static str> = phf_map! {
    "" =>
        ansi!("Type <bold>help <lt>topicname><reset> to learn about a topic. Most \
          commands can be used as a topicname.\n\
          Topics of interest to unregistered users:\n\
          \t<bold>register<reset>\tLearn about the <bold>register<reset> command.\n\
          \t<bold>login<reset>\tLearn how to log in as an existing user.\n"),
    "register" =>
        ansi!("Registers a new user. You are allowed at most 5 at once.\n\
               \t<bold>register <lt>username> <lt>password> <lt>email><reset>\n\
               Email will be used to check you don't have too many accounts and \
               in case you need to reset your password."),
    "login" =>
        ansi!("Logs in as an existing user.\n\
              \t<bold>login <lt>username> <lt>password<reset>")
};

static REGISTERED_HELP_PAGES: phf::Map<&'static str, &'static str> = phf_map! {
    "" =>
        ansi!("Type <bold>help <lt>topicname><reset> to learn about a topic. Most \
          commands can be used as a topicname.\n\
          Topics of interest:\n\
          \t<bold>newbie<reset>\tLearn the absolute basics.\n\
          \t<bold>movement<reset>\tCommands for moving around.\n\
          \t<bold>possessions<reset>\tCommands for dealing with possessions / weapons.\n\
          \t<bold>talk<reset>\tFind out how to talk in the game.\n\
          \t<bold>combat<reset>\tLearn how to fight.\n\
          \t<bold>information<reset>\tLearn how to find out about the world and your character."),
    "newbie" =>
        ansi!("So you've just landed in BlastMud, and want to know how to get started?\n\
               You control your character, and can tell your character to move around the\n\
               world, and see things through their eyes.\n\
               The world (yes, even outside!) is divided up into rooms, and each room has\n
               exits that you are allowed to take, normally called north, south, east, west,\n
               northeast, northwest, southeast, southwest, up and down (sometimes you can also go in).\n\
               \n\
               Try <bold>look<reset> (or <bold>l<reset>) to look at the current room. It will\n\
               also show you all the exits you can take, and a little ASCII-art map showing \n\
               the local layout. <bold>lmap<reset> shows you a more detailed map showing the\n\
               directions you can move in. You can also look at objects or players with \n\
               <bold>l<reset> followed by the name of the object (shortening is okay).\n\
               Once you know what direction to move, you can type the direction, using either\n\
               the full name of the direction (e.g. <bold>northeast<reset> or <bold>south<reset>,\n\
               or a short form (<bold>n<reset>, <bold>e<reset>, <bold>s<reset>, <bold>w<reset>, <bold>ne<reset>, <bold>nw<reset>, <bold>se<reset>, <bold>sw<reset>, <bold>up<reset>, <bold>down<reset>)."),
    "movement" =>
        ansi!("Once you know what direction to move, you can type the direction, using either\n\
               the full name of the direction (e.g. <bold>northeast<reset> or <bold>south<reset>,\n\
               or a short form (<bold>n<reset>, <bold>e<reset>, <bold>s<reset>, <bold>w<reset>, <bold>ne<reset>, <bold>nw<reset>, <bold>se<reset>, <bold>sw<reset>, <bold>up<reset>, <bold>down<reset>)."),
    "look" =>
        ansi!("Try <bold>look<reset> or <bold>l<reset> to look at the current room. Follow it with the \
               name of an exit to look at an adjacent room. Or name an object or player in the room to look \
               at that."),
    "combat" =>
        ansi!("Type <bold>kill<reset> character or <bold>k<reset> character to attack a character. \
               If their health points hits zero, they die. If your health points hit zero, you die. \
               Note that characters can reclone so death isn't fully permanent (see <bold>help death<reset>). \
               Combat depends on your ability to dodge the enemy's attacks (based on your dodge skill), \
               your armour's ability to soak their attacks, and the skill in the weapon you are wielding \
               - and of course the same factors for your opponents. \
               You can only attack one character at a time, but more than one can gang up on you! \
               When you fight, you might notice you improve at your skills (see <bold>help skills<reset>), \
               and also gain more experience. You gain more skills using a weapon matched to your current \
               experience level, and more experience fighting harder opponents - but if you take on someone \
               too tough, you might die and lose experience recloning! \
               Remember, you can always (at least while you are not dead) try to run away from combat by just \
               by using movement commands (see <bold>help movement<reset> - but if your dodge skill isn't high \
               enough, it might fail."),
    "death" =>
        ansi!("In Blastmud, you can die if your health points reach zero. While you are dead, you can't \
               move or talk! Unless someone quickly comes and uses a defibrilator on you, your current \
               body is gone. Luckily, thanks to residual Gazos-Murlison technology that survived the \
               fall of the empire, you can still re-clone into a new body. However, this has some \
               downsides: you lose a bit of your experience in the process, and anything you had on your \
               person stays with the corpse of your old body! So it is certainly worth trying not to die!"),
    "skills" =>
        ansi!(
            "Blastmud has lots of different skills that you can learn. There are two different \
             measures of a skill. Total skill is how good you are in absolute terms - you get some skill from your stats, \
             and some from your raws. Raw skill is how much you have improved by learning the skill as you \
             play the game. Your raw skill caps out at 15 per skill, and adds on to the contribution from \
             your stats, and any temporary buffs or debuffs caused by your character's state."),
    "information" =>
        ansi!(
            "Commands to examine the world:\n\
             \t<bold>look<reset> (or <bold>l<reset>) to look around - follow it with an exit or\n\
             \t\ta character / item name for detail on that.\n\
             \t<bold>lmap<reset> - get a map showing exits and places."),
    "talk" =>
        ansi!("Use:\n\
              \t<bold>'<reset>message to send message to the room.\n\
              \t<bold>-<reset>user message to whisper to someone."),
    "possessions" =>
        ansi!("Use:\n\
              \t<bold>get<reset> item to pick something up from the ground.\n\
              \t<bold>drop<reset> item to drop something from your inventory.\n\
              \t<bold>inv<reset> to see what you are holding.\n\
              \t<bold>buy<reset> item to buy something from the shop you are in.\n\
              \t<bold>list<reset> to see the price list for the stop.\n\
              \t<bold>wield<reset> to hold a weapon in your inventory for use when you attack.\n\
              Hint: get and drop support an item name, but you can also prefix it \
              with a number - e.g. <bold>get 2 whip<reset>. Instead of a number, you can \
              use <bold>all<reset>. You can also omit the item name to match any \
              possession, e.g. <bold>drop all<reset> will drop everything you have.")
};

static EXPLICIT_HELP_PAGES: phf::Map<&'static str, &'static str> = phf_map! {
    "fuck" =>
        ansi!("Type <bold>fuck <lt>name><reset> to fuck someone. It only works if \
               they have consented.")
};

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(self: &Self, ctx: &mut VerbContext, _verb: &str, remaining: &str) -> UResult<()> {
        let mut help = None;
        let is_unregistered = match ctx.user_dat {
            None => true,
            Some(user_dat) => !user_dat.terms.terms_complete
        };
        if is_unregistered {
            help = help.or_else(|| UNREGISTERED_HELP_PAGES.get(remaining));
        } else {
            help = help.or_else(|| REGISTERED_HELP_PAGES.get(remaining));
            if !ctx.session_dat.less_explicit_mode {
                help = help.or_else(|| EXPLICIT_HELP_PAGES.get(remaining))
            }
        }
        help = help.or_else(|| ALWAYS_HELP_PAGES.get(remaining));
        let help_final = help.ok_or(
            UserError("No help available on that".to_string()))?;
        ctx.trans.queue_for_session(ctx.session,
                                    Some(&(help_final.to_string() + "\n"))
        ).await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
