use serde::{Serialize, Deserialize};
use crate::{
    models::item::{SkillType, Item, Pronouns}
};
use once_cell::sync::OnceCell;
use std::collections::BTreeMap;
use rand::seq::SliceRandom;
use super::species::BodyPart;

pub type AttackMessageChoice = Vec<Box<dyn Fn(&Item, &Item, bool) -> String + 'static + Sync + Send>>;
pub type AttackMessageChoicePart = Vec<Box<dyn Fn(&Item, &Item, &BodyPart, bool) -> String + 'static + Sync + Send>>;

pub struct SkillScaling {
    pub skill: SkillType,
    pub min_skill: f64,
    pub mean_damage_per_point_over_min: f64
}

pub struct WeaponData {
    pub uses_skill: SkillType,
    pub raw_min_to_learn: f64,
    pub raw_max_to_learn: f64,
    pub normal_attack_start_messages: AttackMessageChoice,
    pub normal_attack_success_messages: AttackMessageChoicePart,
    pub normal_attack_mean_damage: f64,
    pub normal_attack_stdev_damage: f64,
    pub normal_attack_skill_scaling: Vec<SkillScaling>,
}

impl Default for WeaponData {
    fn default() -> Self {
        Self {
            uses_skill: SkillType::Blades,
            raw_min_to_learn: 0.0,
            raw_max_to_learn: 15.0,
            normal_attack_start_messages:
            vec!(Box::new(|attacker, victim, exp| format!(
                "{} makes an attack on {}",
                &attacker.display_for_sentence(exp, 1, true),
                &victim.display_for_sentence(exp,1, false)))),
            normal_attack_success_messages:
            vec!(Box::new(|attacker, victim, part, exp|
                          format!("{}'s attack on {} hits {} {}",
                                  &attacker.display_for_sentence(exp, 1, true),
                                  &victim.display_for_sentence(exp, 1, false),
                                  &victim.pronouns.possessive,
                                  part.display(victim.sex.clone())
                          ))),
            normal_attack_mean_damage: 1.0,
            normal_attack_stdev_damage: 2.0,
            normal_attack_skill_scaling: vec!()
        }
    }
}

pub struct ChargeData {
    pub max_charges: u8,
    pub charge_name_prefix: &'static str,
    pub charge_name_suffix: &'static str,
}

impl Default for ChargeData {
    fn default() -> Self {
        Self {
            max_charges: 1,
            charge_name_prefix: "charge",
            charge_name_suffix: "",
        }
    }
}

pub enum UseEffect {
    // messagef takes player, item used, target as the 3 parameters. Returns (explicit, non explicit) message.
    BroadcastMessage { messagef: Box<dyn Fn(&Item, &Item, &Item) -> (String, String) + Sync + Send>},
    // skill_multiplier is always positive - sign flipped for crit fails.
    ChangeTargetHealth { delay_secs: u64, base_effect: i64, skill_multiplier: f64,
                         max_effect: i64, message: Box<dyn Fn(&Item) -> (String, String) + Sync + Send> }
}

pub struct UseData {
    pub uses_skill: SkillType,
    pub diff_level: f64,
    pub crit_fail_effects: Vec<UseEffect>,
    pub fail_effects: Vec<UseEffect>,
    pub success_effects: Vec<UseEffect>,
    pub errorf: Box<dyn Fn(&Item, &Item) -> Option<String> + Sync + Send>,
    pub task_ref: &'static str,    
}

impl Default for UseData {
    fn default() -> Self {
        Self {
            uses_skill: SkillType::Medic,
            diff_level: 10.0,
            crit_fail_effects: vec!(),
            fail_effects: vec!(),
            success_effects: vec!(),
            errorf: Box::new(|_it, _target| None),
            task_ref: "set me",
        }
    }
}

pub struct PossessionData {
    pub weapon_data: Option<WeaponData>,
    pub display: &'static str,
    pub display_less_explicit: Option<&'static str>,
    pub details: &'static str, 
    pub details_less_explicit: Option<&'static str>,
    pub aliases: Vec<&'static str>,
    pub max_health: u64,
    pub charge_data: Option<ChargeData>,
    pub use_data: Option<UseData>,
    pub weight: u64,
}

impl Default for PossessionData {
    fn default() -> Self {
        Self {
            weapon_data: None,
            display: "Thingy",
            display_less_explicit: None,
            details: "A generic looking thing",
            details_less_explicit: None,
            aliases: vec!(),
            max_health: 10,
            weight: 100,
            charge_data: None,
            use_data: None,
        }
    }
}

impl WeaponData {
    pub fn normal_attack_start_message(
        &self,
        attacker: &Item, victim: &Item, explicit_ok: bool) -> String {
        let mut rng = rand::thread_rng();
        self.normal_attack_start_messages[..].choose(&mut rng).map(
            |f| f(attacker, victim, explicit_ok)).unwrap_or(
            "No message defined yet".to_owned())
    }

    pub fn normal_attack_success_message(
        &self, attacker: &Item, victim: &Item,
        part: &BodyPart, explicit_ok: bool
    ) -> String {
        let mut rng = rand::thread_rng();
        self.normal_attack_success_messages[..].choose(&mut rng).map(
            |f| f(attacker, victim, part, explicit_ok)).unwrap_or(
            "No message defined yet".to_owned())
    }

    
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum PossessionType {
    // Special values that substitute for possessions.
    Fangs, // Default weapon for certain animals
    // Real possessions from here on:
    AntennaWhip,
    MediumTraumaKit,
}

impl Into<Item> for PossessionType {
    fn into(self) -> Item {
        let possession_dat = possession_data().get(&self).unwrap();
        Item {
            item_type: "possession".to_owned(),
            possession_type: Some(self.clone()),
            display: possession_dat.display.to_owned(),
            display_less_explicit: possession_dat.display_less_explicit.map(|d| d.to_owned()),
            details: Some(possession_dat.details.to_owned()),
            details_less_explicit: possession_dat.details_less_explicit.map(|d| d.to_owned()),
            aliases: possession_dat.aliases.iter().map(|al| (*al).to_owned()).collect(),
            health: possession_dat.max_health,
            weight: possession_dat.weight,
            pronouns: Pronouns {
                is_proper: false,
                ..Pronouns::default_inanimate()
            },
            charges: possession_dat.charge_data.as_ref()
                .map(|cd| cd.max_charges).unwrap_or(0),
            ..Default::default()
        }
    }
}

pub fn fist() -> &'static WeaponData {
    static FIST_WEAPON: OnceCell<WeaponData> = OnceCell::new();
    FIST_WEAPON.get_or_init(|| {
        WeaponData {
            uses_skill: SkillType::Fists,
            raw_min_to_learn: 0.0,
            raw_max_to_learn: 2.0,
            normal_attack_start_messages: vec!(
                Box::new(|attacker, victim, exp|
                         format!("{} swings at {} with {} fists",
                                 &attacker.display_for_sentence(exp, 1, true),
                                 &victim.display_for_sentence(exp, 1, false),
                                 &attacker.pronouns.possessive
                         )
                )
            ),
            normal_attack_success_messages: vec!(
                Box::new(|attacker, victim, part, exp|
                         format!("{}'s fists smash into {}'s {}",
                                 &attacker.display_for_sentence(exp, 1, true),
                                 &victim.display_for_sentence(exp, 1, false),
                                 &part.display(victim.sex.clone())
                         )
                )
            ),
            ..Default::default()
        }
    })
}

pub fn possession_data() -> &'static BTreeMap<PossessionType, PossessionData> {
    static POSSESSION_DATA: OnceCell<BTreeMap<PossessionType, PossessionData>> = OnceCell::new();
    use PossessionType::*;
    &POSSESSION_DATA.get_or_init(|| {
        vec!(
            (Fangs, PossessionData {
                weapon_data: Some(WeaponData {
                    uses_skill: SkillType::Fists,
                    raw_min_to_learn: 0.0,
                    raw_max_to_learn: 2.0,
                    normal_attack_start_messages: vec!(
                        Box::new(|attacker, victim, exp|
                                 format!("{} bares {} teeth and lunges at {}",
                                         &attacker.display_for_sentence(exp, 1, true),
                                         &attacker.pronouns.possessive,
                                         &victim.display_for_sentence(exp, 1, false),
                                 )
                        )
                    ),
                    normal_attack_success_messages: vec!(
                        Box::new(|attacker, victim, part, exp|
                                 format!("{}'s teeth connect and tear at the flesh of {}'s {}",
                                         &attacker.display_for_sentence(exp, 1, true),
                                         &victim.display_for_sentence(exp, 1, false),
                                         &part.display(victim.sex.clone())
                                 )
                        )
                    ),
                    ..Default::default()
                }),
                ..Default::default()
            }),
            (AntennaWhip, PossessionData {
                display: "antenna whip",
                details: "A crudely fashioned whip made from a broken metal antenna. It looks a bit flimsy, but it \
                          might do you until you get a better weapon!",
                aliases: vec!("whip"),
                weapon_data: Some(WeaponData {
                    uses_skill: SkillType::Whips,
                    raw_min_to_learn: 0.0,
                    raw_max_to_learn: 2.0,
                    normal_attack_start_messages: vec!(
                        Box::new(|attacker, victim, exp|
                                 format!("{} lines up {} antenna whip for a strike on {}",
                                         &attacker.display_for_sentence(exp, 1, true),
                                         &attacker.pronouns.possessive,
                                         &victim.display_for_sentence(exp, 1, false),
                                 )
                        )
                    ),
                    normal_attack_success_messages: vec!(
                        Box::new(|attacker, victim, part, exp|
                                 format!("{}'s antenna whip scores a painful red line across {}'s {}",
                                         &attacker.display_for_sentence(exp, 1, true),
                                         &victim.display_for_sentence(exp, 1, false),
                                         &part.display(victim.sex.clone())
                                 )
                        )
                    ),
                    normal_attack_mean_damage: 3.0,
                    normal_attack_stdev_damage: 3.0,
                    ..Default::default()
                }),
                ..Default::default()
            }),
            (MediumTraumaKit, PossessionData {
                display: "medium trauma kit",
                details: "A collection of bandages and and small gadgets that look like they could, in the right hands, make an unhealthy person healthy again. It looks like when brand new, it could be used 5 times.",
                aliases: vec!("trauma"),
                charge_data: Some(ChargeData {
                    max_charges: 5,
                    charge_name_prefix: "treatment",
                    charge_name_suffix: "worth of supplies",
                    ..Default::default()
                }),
                use_data: Some(UseData {
                    uses_skill: SkillType::Medic,
                    diff_level: 10.0,
                    crit_fail_effects: vec!(
                        UseEffect::BroadcastMessage {
                          messagef: Box::new(|player, _item, target| (
                              format!(
                                  "{} attempts to heal {} with a trauma kit, but fucks it up badly\n",
                                  &player.display_for_sentence(true, 1, true),
                                  &if target.item_type == player.item_type && target.item_code == player.item_code {
                                      player.pronouns.intensive.clone()
                                  } else {
                                      target.display_for_sentence(true, 1, false)
                                  }
                              ),
                              format!("{} attempts to heal {} with a trauma kit, but messes it up badly\n",
                                  &player.display_for_sentence(false, 1, true),
                                  &if target.item_type == player.item_type && target.item_code == player.item_code {
                                      player.pronouns.intensive.clone()
                                  } else {
                                      target.display_for_sentence(false, 1, false)
                                  }
                              )))
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 0, base_effect: -2, skill_multiplier: -3.0,
                            max_effect: -5,
                            message: Box::new(
                                |target|
                                (format!(
                                    "Fuck! The trauma kit makes {}'s condition worse",
                                    target.display_for_sentence(true, 1, false)),
                                 format!(
                                     "The trauma kit makes {}'s condition worse",
                                     target.display_for_sentence(false, 1, false)
                                 )
                                ))
                        }
                    ),
                    fail_effects: vec!(
                        UseEffect::BroadcastMessage {
                          messagef: Box::new(|player, _item, target| (
                              format!(
                                  "{} attempts unsuccessfully to heal {} with a trauma kit\n",
                                  &player.display_for_sentence(true, 1, true),
                                  &if target.item_type == player.item_type && target.item_code == player.item_code {
                                      player.pronouns.intensive.clone()
                                  } else {
                                      target.display_for_sentence(true, 1, false)
                                  }
                              ),
                              format!("{} attempts unsuccessfully to heal {} with a trauma kit\n",
                                  &player.display_for_sentence(false, 1, true),
                                  &if target.item_type == player.item_type && target.item_code == player.item_code {
                                      player.pronouns.intensive.clone()
                                  } else {
                                      target.display_for_sentence(false, 1, false)
                                  }
                              )))
                        },
                    ),
                    success_effects: vec!(
                        UseEffect::BroadcastMessage {
                          messagef: Box::new(|player, _item, target| (
                              format!(
                                  "{} expertly heals {} with a trauma kit\n",
                                  &player.display_for_sentence(true, 1, true),
                                  &if target.item_type == player.item_type && target.item_code == player.item_code {
                                      player.pronouns.intensive.clone()
                                  } else {
                                      target.display_for_sentence(true, 1, false)
                                  }
                              ),
                              format!("{} expertly heals {} with a trauma kit\n",
                                  &player.display_for_sentence(false, 1, true),
                                  &if target.item_type == player.item_type && target.item_code == player.item_code {
                                      player.pronouns.intensive.clone()
                                  } else {
                                      target.display_for_sentence(false, 1, false)
                                  }
                              )))
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 0, base_effect: 2, skill_multiplier: 8.0,
                            max_effect: 10,
                            message: Box::new(
                                |target|
                                (format!(
                                    "FUUUCK! It hurts {}, but also starts to soothe {}",
                                    target.display_for_sentence(true, 1, false),
                                    &target.pronouns.object
                                ),
                                 format!(
                                     "It hurts {}, but also starts to soothe {}",
                                     target.display_for_sentence(true, 1, false),
                                     &target.pronouns.object
                                 ))
                            )
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 10, base_effect: 2, skill_multiplier: 7.0,
                            max_effect: 9,
                            message: Box::new(
                                |target|
                                (format!(
                                    "FUUUCK! It hurts {}, but also starts to soothe {}",
                                    target.display_for_sentence(true, 1, false),
                                    &target.pronouns.object
                                ),
                                 format!(
                                     "It hurts {}, but also starts to soothe {}",
                                     target.display_for_sentence(true, 1, false),
                                     &target.pronouns.object
                                 ))
                            )
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 20, base_effect: 1, skill_multiplier: 6.0,
                            max_effect: 7,
                            message: Box::new(
                                |target|
                                (format!(
                                    "The bandages soothe {}'s wounds",
                                    target.display_for_sentence(true, 1, false),
                                ),
                                 format!(
                                    "The bandages soothe {}'s wounds",
                                     target.display_for_sentence(false, 1, false),
                                 ))
                            )
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 30, base_effect: 1, skill_multiplier: 5.0,
                            max_effect: 6,
                            message: Box::new(
                                |target|
                                (format!(
                                    "The bandages soothe {}'s wounds",
                                    target.display_for_sentence(true, 1, false),
                                ),
                                 format!(
                                    "The bandages soothe {}'s wounds",
                                     target.display_for_sentence(false, 1, false),
                                 ))
                            )
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 40, base_effect: 0, skill_multiplier: 4.0,
                            max_effect: 4,
                            message: Box::new(
                                |target|
                                (format!(
                                    "The bandages soothe {}'s wounds",
                                    target.display_for_sentence(true, 1, false),
                                ),
                                 format!(
                                    "The bandages soothe {}'s wounds",
                                     target.display_for_sentence(false, 1, false),
                                 ))
                            )
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 50, base_effect: 0, skill_multiplier: 3.0,
                            max_effect: 3,
                            message: Box::new(
                                |target|
                                (format!(
                                    "The bandages soothe {}'s wounds",
                                    target.display_for_sentence(true, 1, false),
                                ),
                                 format!(
                                    "The bandages soothe {}'s wounds",
                                     target.display_for_sentence(false, 1, false),
                                 ))
                            )
                        },
                        UseEffect::ChangeTargetHealth {
                            delay_secs: 60, base_effect: 0, skill_multiplier: 2.0,
                            max_effect: 2,
                            message: Box::new(
                                |target|
                                (format!(
                                    "The bandages soothe {}'s wounds",
                                    target.display_for_sentence(true, 1, false),
                                ),
                                 format!(
                                    "The bandages soothe {}'s wounds",
                                     target.display_for_sentence(false, 1, false),
                                 ))
                            )
                        },
                    ),
                    task_ref: "bandage",
                    errorf: Box::new(
                        |_item, target|
                        if target.is_dead {
                            Some(format!("It is too late, {}'s dead", target.pronouns.subject))
                        } else if target.item_type != "player" && target.item_type != "npc" {
                            Some("It only works on animals.".to_owned())
                        } else {
                            None
                        }),
                    ..Default::default()
                }),
                ..Default::default()
            }),
        ).into_iter().collect()
    })
}
